
/*
jppp18 (c) 2010-always Jan Panteltje

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


                                            /* set tabs to 4 to read this source */


#include "jppp18.h"

extern int flash_buffer_flags[];

typedef unsigned int  word;
typedef unsigned char byte, bit;


// *********************************************************************
// HEX FILE LOADER
// *********************************************************************

// This is for Intel INHX8M (8-bit merged) hex files only.
// These files use two bytes for each word (low, then high).
// All addresses are doubled, i.e., 0x2001 is encoded as 0x4002,
// so that addresses increment at the same rate as the byte count.


/* stuffs count data bytes into array array at address - base from array data, tests for > size, sets *used, returns true if successful. */
int stuff_array(int address, int8_t *array, int base, int size, int *used, int8_t *data, int count)
{
int i;

if(debug_flag)
	{
	fprintf(stderr, "stuff_array(): arg address=0x%04x, array=%p base=%0x04x size=%d used=%p data=%p count=%d\n", address, array, base, size, used, data, count); 
	}

//fprintf(stderr, "address: %04XH base=%04x size=%d *used=%d count=%d\n", address+count-1, base, size, *used, count);

if(address - base + count - 1 > size)
	{
	printf("Invalid address: %04XH\n", address + count - 1);

	return 0;
	}

for(i = 0; i < count; i++)
	{
	array[i + address - base] = data[i];

	if(*used < address - base + count) *used = address - base + count;
	}

return 1;
} /* end function stuff_array */


int valid_hex_line(char *s)				// Gross syntax and checksum check.
{
byte cksum = 0;                     	// For all HEX formats, not just 8M.
int bytecount;
int i, b;

if (s[0] != ':')
	{
	fprintf(stderr, "initial colon missiing.\n");
	
	return 0;          	// Initial colon
	}
	
sscanf(s + 1, "%2x", &bytecount);
if (bytecount > 32)
	{
	fprintf(stderr, "invalid byte count.\n");
	return 0;       	// Valid byte count
	}
cksum = bytecount;
i = 3;
bytecount = bytecount + 3;

while (bytecount > 0)
	{
	bytecount--;
	sscanf(s + i, "%2x", &b);
	cksum = cksum + b;                  // Compute checksum
	i = i + 2;
	}

sscanf(s + i, "%2x", &b);

cksum =- cksum;
if (cksum == b) return 1;            	// Test checksum

fprintf(stderr, "checksum error.\n");
 
return 0;
} /* end function valid_hex_line */


// Loads a hex file into memory arrays
int load_hex_file(char *filename)
{
char s[256];
word i;
word linetype = 0;   // 0 for data, 1 for end of file
word wordcount;      // number of 16 bit words on this line
word address;        // address where they begin
int line;
int position;
unsigned int ua;
int upper_linear_base_address;
int8_t data[16];	// 16 bytes max per line of hex
FILE *fptr;

if(debug_flag)
	{	
	fprintf(stderr, "load_hex_file(): arg filename=%s\n", filename);
	}

if(! filename) return 0;

fptr = fopen(filename, "r");
if(! fptr)
	{
	fprintf(stderr, "load_hex_file(): could not open file %s for read.\n", filename);
          
	return 0;
	}

fprintf(stderr, "Loading hex file:\n");

/* clear sizes */
eeprom_used = 0;
config_used = 0;
id_used = 0;
program_used = 0;

line = 0;
upper_linear_base_address = 0;
while((!feof(fptr)) && (linetype != 1))
	{
    fgets(s, 255, fptr);
	line++;

	if(debug_flag)
		{
		s[80] = 0;
		fprintf(stderr, "load_hex_file: line=%d reads=%s\n", line, s);
		}

	if(!valid_hex_line(s)) 
		{			// Syntax check
		s[80] = 0;	// Truncate invalid line for display  was 40

		if (s[0] != ':')
			{
       		fprintf(stderr, "jppp18: invalid line (skipped): '%s'...\n", s);
       	 	continue;
     		}
		else
			{
    	    fprintf(stderr, "jppp18: unable to decode line:  '%s'...\n", s);


			return 0;
			}
		}

	sscanf(s + 1, "%2x", &wordcount);      // Parse the line - Intel Hex8M

	sscanf(s + 3, "%4x", &address);

	sscanf(s + 7, "%2x", &linetype);

	if(linetype == 1)
		{
//		fprintf(stderr, "Loaded:\n");
		fprintf(stderr, " Program %d bytes at address 0x%06x\n", program_used, program_base);
		fprintf(stderr, " ID      %d bytes at address 0x%06x\n", id_used, id_base);
		fprintf(stderr, " Config  %d bytes at address 0x%06x\n", config_used, config_base);
		fprintf(stderr, " EEPROM  %d bytes at address 0x%06x\n", eeprom_used, eeprom_base);

		return 1;
		}
	
	if(linetype == 4) /* extended linear address record */
		{
		sscanf(s + 9, "%4x", &upper_linear_base_address); 		

		if(debug_flag)
			{
			fprintf(stderr, "load_hex_file(): line %d extended linear hex record detected\n", line);		
			/* the first 2 databytes afer the linetype set a ULBA base address to be added to address */
			fprintf(stderr, "load_hex_file(): line %d set upper_linear_base_address=0x%04x\n", line, upper_linear_base_address);
			}

		continue;		
		}

	if(debug_flag)
		{
		fprintf(stderr, "jppp18: hex line address=0x%04x (%d)\n", address, address);
		}
		
	/* grab the data */
	for (i = 0; i < wordcount; i++)
		{
		sscanf(s + (9 + (2 * i) ), "%2x", &ua);
		data[i] = ua;
		}

	/* add upper_linear_base_address */
	position = address + (upper_linear_base_address << 16);

	/* Place in appropriate array */
	if(position >= eeprom_base)
		{
		/* stuff count data bytes into array array at address - base from array data, tests for > size, sets *used, returns true if successful. */
		// int stuff_array(int address, int8_t array[], int base, int size, int *used, int8_t data[],  int count)
		if(! stuff_array(position, eeprom_data, eeprom_base, eeprom_size, &eeprom_used, data, wordcount) ) return 0;
		}
	else if(position >= config_base)
		{
		if(! stuff_array(position, config_data, config_base, config_size, &config_used, data, wordcount) ) return 0;
		}
	else if(position >= id_base)
		{
		if(! stuff_array(position, id_data, id_base, id_size, &id_used, data, wordcount) ) return 0;
		}
	else
		{
		if(!stuff_array(position, program_data, program_base, program_size, &program_used, data, wordcount) ) return 0;
		}

	} /* end while data */

return 1;
} /* end function load_hex_file */

