
/*
jppp18 (c) 2010-always Jan Panteltje

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



                                            /* set tabs to 4 to read this source */

#include "jppp18.h"


// write a hex segment specifying start at base_address, using length bytes from buffer[0] upwards
int write_hex_section(FILE *fptr, int base_address, int length, int8_t *buffer)
{
int i;
int record_type;
int position;
int space;
int record_length;
int checksum;

if(debug_flag)
	{
	fprintf(stderr, "write_hex_section(): base_address=0x%06x length=%d buffer=%p\n", base_address, length, buffer);
	}

if(! fptr) return 0;
if(! buffer) return 0;
if( base_address < 0) return 0;
if(! buffer) return 0;

/* if base_addres is bigger then 16 bits can hold, then first write an extended Intel hex record specifying the base address */
if(base_address > 0xffff)
	{
	/* : 02 0000 04 0030 CA */

	/* recod length is always 2 for thid type of line */
	record_length = 2;

	/* calculate checksum */
	checksum = 0 - (record_length + 0x00 + 0x00 + 0x04 + ( (base_address >> 16) & 0xff) + ( (base_address >> 24) & 0xff) );

	/* print extended Intel hex record */
	fprintf(fptr, ":%02x%04x%02x%04x%02x\n", record_length, 0x0000, 0x04, base_address >> 16, checksum & 0xff);
	}

record_type = 0;
position = 0;
space = length;
while(1)
	{
	/* write a hex line, the format is:  ':' length(1), address(2), recordtype(1) RECORD_LENGTH data bytes(1), checksum(1) */
	/* :10 0020 00 00 F0 D0 80 00 01 33 84 0F 01 D0 B6 1B EF 00 F0 48 */

	checksum = 0;

	/* calculate length of record */
	if(space < RECORD_LENGTH) record_length = space;
	else record_length = RECORD_LENGTH;

	/* print line header */
	fprintf(fptr, ":%02x%04x%02x", record_length, position, record_type); 

	checksum += record_length;

	checksum += (position >> 8) & 0xff;
	checksum += position & 0xff;

	/* print line data bytes */
	for(i = 0; i < record_length; i++)
		{
		fprintf(fptr, "%02x", buffer[position + i] & 0xff);
	
		checksum += buffer[position + i] & 0xff;
		} /* end for each data byte in a recoed */

	/* print line checksum */
	/* The sum of all the ASCII pairs in a record after converting to binary, from the RECLEN including the CHKSUM field, is zero. */
	fprintf(fptr, "%02x\n", (0 - (checksum &0xff) ) & 0xff );

	space -= record_length;
	if(space == 0) break;

	position += record_length;
	} /* end wile print hex record lines */

return 1;
} /* end function write_hex_section */



// write Intel format hex file from program space, ID space, config space, and EEPROM space
int write_hex_file(char * filename)
{
FILE *fptr;
char hex_filename[4096];
char *ptr;

if(debug_flag)
	{
	fprintf(stderr, "hex_out(): arg filename=%s\n", filename);
	}

if(! filename) return 0;

/* append .hex if not .hex at end of filename */
strcpy(hex_filename, filename);
if(strstr(hex_filename, ".hex") != (hex_filename + strlen(hex_filename) - 4) ) strcat(hex_filename, ".hex");

fptr = fopen(hex_filename, "r");
if(fptr)
	{
	fclose(fptr);

	fprintf(stderr, "jppp18: write_hex_file(): file %s already exists, overwrite (y/n)?\n", hex_filename);
	fscanf(stdin, "%as", &ptr);

	if( toupper(ptr[0]) != 'Y')
		{
		fprintf(stderr, "write hex file cancelled.\n");		
		
		return 0;
		}
	} /* end if file already exists */

fptr = fopen(hex_filename, "w");
if(! fptr)
	{
	fprintf(stderr, "write_hex_file(): could not open file %s for write, command cancelled.\n", hex_filename);
				
	return 0;
	}

fprintf(stderr, "saving to file %s\n", hex_filename);

/* program space */
fprintf(stderr, "reading program space.\n");
if(! read_chip(program_base, program_size, flash_read_buffer) )
	{
	fprintf(stderr, "read_chip failed.\n");

	fclose(fptr);

	return 0;
	}

fprintf(stderr, "writing program space to %s.\n", hex_filename);

//int write_hex_section(FILE* fptr, int address, int length, int8_t *buffer)
if(! write_hex_section(fptr, program_base, program_size, flash_read_buffer) )
	{
	fprintf(stderr, "could not write hex section.\n");

	fclose(fptr);

	return 0;
	} 


/* ID space */
fprintf(stderr, "reading ID space.\n");
if(! read_chip(id_base, id_size, flash_read_buffer) )
	{
	fprintf(stderr, "read_chip failed.\n");

	fclose(fptr);

	return 0;
	}

fprintf(stderr, "appending ID space to %s.\n", hex_filename);

//int write_hex_section(FILE* fptr, int address, int length, int8_t *buffer)
if(! write_hex_section(fptr, id_base, id_size, flash_read_buffer) )
	{
	fprintf(stderr, "could not write hex section.\n");

	fclose(fptr);

	return 0;
	} 


/* config space */
fprintf(stderr, "reading config space.\n");
if(! read_chip(config_base, config_size, flash_read_buffer) )
	{
	fprintf(stderr, "read_chip failed.\n");

	fclose(fptr);

	return 0;
	}

fprintf(stderr, "appending config space to %s.\n", hex_filename);

//int write_hex_section(FILE* fptr, int address, int length, int8_t *buffer)
if(! write_hex_section(fptr, config_base, config_size, flash_read_buffer) )
	{
	fprintf(stderr, "could not write hex section.\n");

	fclose(fptr);

	return 0;
	} 


/* EEPROM space */
fprintf(stderr, "reading EEPROM space.\n");
if(! read_eeprom(0, eeprom_size, eeprom_read_buffer) )
	{
	fprintf(stderr, "read_chip failed.\n");

	fclose(fptr);
	
	return 0;
	}

fprintf(stderr, "appending EEPROM space to %s.\n", hex_filename);

//int write_hex_section(FILE* fptr, int address, int length, int8_t *buffer)
if(! write_hex_section(fptr, eeprom_base, eeprom_size, eeprom_read_buffer) )
	{
	fprintf(stderr, "could not write hex section.\n");

	fclose(fptr);

	return 0;
	} 

/* write end of file record */
fprintf(fptr, ":00000001FF\n");
 
fclose(fptr);

return 1;
} /* end function hex_out */


