
/*
jppp18 (c) 2010-always Jan Panteltje

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



                                            /* set tabs to 4 to read this source */



/* I/O routines for noppp programmer hardware via parport */

#include "jppp18.h"

// *********************************************************************
// PARALLEL PORT HARDWARE INTERFACE
// *********************************************************************

// Uses two-wire serial data communication through printer port.
// Pin 1, STROBE, is serial clock;
// Pin 14, AUTOFD, is serial data out to PIC;
// Pin 11, BUSY, is serial data in from PIC;
// Pin 17, SLCTIN, is low when writing, high to provide pull-up when reading;
// Pin 2, D0, is lowered to apply Vpp.
// Pin 15, !ERRORis lowered to set Vcc to 13 V (bit 3 0x379)
// Pin 3, D1		(bit D1 0x378 data)
#define VCC_CONTROL	1
#define VPP_CONTROL	0

// SLCTIN and BUSY are tied together for pull-up and for hardware detection.
// (In current versions this is done through diodes or gates.)

// SLCTIN is an open-collector output with pull-up.
// If it is pulled down, some printer ports will latch it down.
// Accordingly, it and all the other control bits are asserted 
// every time they are needed.


// data

void pgdl()
{
int a;

a = inb(parport_control);
a |= 0x02;					 // AUTOFD down		pin 14
outb(a, parport_control);
}


void pgdh()
{
int a;

a = inb(parport_control);
a &= ~0x02;					// AUTOFD up		pin 14
outb(a, parport_control);
}


// clock

void pgcl()
{
int a;

a = inb(parport_control);
a |= 0x01;					// STROBE down		pin 1
outb(a, parport_control);
}


void pgch()
{
int a;

a = inb(parport_control);
a &= ~0x01;					// STROBE up	p	in 1
outb(a, parport_control);
}



// data in

int pgdin()
{


return ( (~inb(parport_status)) & 0x80) >> 7;	// BUSY pin 11
}
 


// Vdd 

void vdd5()
{
int a;

a = inb(parport_data);
a &= ~32;						/* reset bit 5 (d5 on pin 7) */
outb(a, parport_data);
}
 

void vdd0()
{
int a;
a = inb(parport_data);
a |= 32;						/* set bit 5 (d5 on pin 7) */
outb(a, parport_data);
}


// Vpp

void vpp0() // INIT up, D0 up
{
int a;

a = inb(parport_data);
a |= 0x01;						/* set bit 0 (d0 on pin 2) */
outb(a, parport_data);
}


void vpp9()
{
int a;

a = inb(parport_data);
a &= 0xfe;						/* reset bit 0 (d1 on pin 2) */
outb(a, parport_data);
}



void allpinslow()
{
int a;

vpp0();

datawritable(); 

pgdl();
pgcl();
a = 0x0F;
outb(parport_data, a);
}


int detecthardware()		// True if BUSY and SLCTIN are tied together.
{
datawritable();				// SLCTIN down
pgdh();						// AUTOFD up

delay(10);

if (pgdin() == 1)
	{
	fprintf(stdout, "BUSY was high after SLCTIN low and AUTOFEED high\n");

	return 0;
	}

datareadable();				// SLCTIN up
pgdh();						// AUTOFD up
delay(10);

if (pgdin() == 0)
	{
	fprintf(stdout, "BUSY was low after SLCTIN high and AUTOFEED high\n");

	return 0;
	}

return 1;
} /* end function detect hardware */



void datawritable()
{
int a;

a = inb(parport_control);
a |= 0x0A;					// SLCTIN pin 17, 	AUTOFD down pin 14
outb(a, parport_control); 
}


void datareadable()
{
int a;

a = inb(parport_control);
a &= ~0x0A;					// SLCTIN pin 17,	 AUTOFD up pin 14
outb(a, parport_control);
}




