
/*
jppp18 (c) 2010-always Jan Panteltje

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include "jppp18.h"




char *strsave(char *s) /* save char array s somewhere */
{
char *p;

p = (char *) malloc(strlen(s) +  1);
if(p) strcpy(p, s);
return p;
} /* end function strsave */



int pic_select(char *pic_type)
{
if(debug_flag)
	{
	fprintf(stderr, "pic_select(): arg pic_typ=%s\n", pic_type);
	}

if(! pic_type) return 0;

/* think we could parse the .inc file[s] for this if present */

/* select a pic */
if(strcmp(pic_type, "18F14K22") == 0)
	{
	// program FLASH
	program_base = 0;
	program_size = 0x4000;

	// id_data;      
	id_base = 0x200000;
	id_size = 8;

	// config_data;
	config_base = 0x300000;
	config_size = 14;

	// eeprom_data;
	eeprom_base = 0xf00000;
	eeprom_size = 256;  

	// device ID data
	devid_base = 0x3ffffe;
	devid_size = 2;
	}
#ifdef MORE_PICS
/* add an entry like this for each supported PIC 18F, you will find the related parameters in the .INC include file for that PIC */
else if(strcmp(pic_type, "18FXXXXX") == 0)
	{
	// program FLASH
	program_base = 0;
	program_size = 0;

	// id_data;      
	id_base = o;
	id_size = 0;

	// config_data;
	config_base = 0;
	config_size = 0;

	// eeprom_data;
	eeprom_base = 0;
	eeprom_size = 0;  

	// device ID data
	devid_base = 0;
	devid_size = 0;
	}
#endif // MORE_PICS
else
	{
	/* unknown PIC type */
	return 0;
	}

if(program_data) free(program_data);
program_data = calloc(program_size, 1); 
if(! program_data)
	{
	fprintf(stderr, "jppp18: pic_select(): could not allocate memory for program_data.\n");

	return 0;
	}	

if(config_data) free(config_data);
config_data = calloc(config_size, 1); 
if(! config_data)
	{
	fprintf(stderr, "jppp18: pic_select(): could not allocate memory for config_data.\n");

	return 0;
	}	

if(id_data) free(id_data);
id_data = calloc(id_size, 1); 
if(! id_data)
	{
	fprintf(stderr, "jppp18: pic_select(): could not allocate memory for id_data.\n");

	return 0;
	}	

if(eeprom_data) free(eeprom_data);
eeprom_data = calloc(eeprom_size, 1); 
if(! eeprom_data)
	{
	fprintf(stderr, "jppp18: pic_select(): could not allocate memory for eeprom_data.\n");

	return 0;
	}	

if(devid_data) free(devid_data);
devid_data = calloc(devid_size, 1);
if(! devid_data)
	{
	fprintf(stderr, "jppp18: pic_select(): could not allocate memory for devid_data.\n");		
	
	return 0;
	}

if(flash_read_buffer) free(flash_read_buffer);
flash_read_buffer = calloc(program_size, 1);
if(! flash_read_buffer)
	{
	fprintf(stderr, "jppp18: pic_select(): could not allocate memory for flash_read_buffer.\n");
	
	return 0;
	}

if(eeprom_read_buffer) free(eeprom_read_buffer);
eeprom_read_buffer = calloc(eeprom_size, 1);
if(! eeprom_read_buffer)
	{
	fprintf(stderr, "jppp18: pic_select(): could not allocate memory for eeprom_read_buffer.\n");

	return 0;	
	}

return 1;
} /* end function pic select */



void io_test()
{
char temp[512];

if(debug_flag)
	{
	fprintf(stderr, "io_test, arg none\n");
	}

allpinslow();

while(1)
	{
	fprintf(stderr, "Vdd (0/5/-1)?\n");
	fscanf(stdin, "%s", temp);

	if(atoi(temp) == 0) vdd0();
	if(atoi(temp) == 5) vdd5();
	if(atoi(temp) == -1) break;
	}

while(1)
	{
	fprintf(stderr, "Vpp (0/12/-1)?\n");
	fscanf(stdin, "%s", temp);

	if(atoi(temp) == 0) vpp0();
	if(atoi(temp) == 12) vpp9();

	if(atoi(temp) == -1) break;
	}

while(1)
	{
	fprintf(stderr, "pgc (0/1/-1)?\n");
	fscanf(stdin, "%s", temp);

	if(atoi(temp) == 0) pgcl();
	if(atoi(temp) == 1) pgch();
	if(atoi(temp) == -1) break;
	}

while(1)
	{
	fprintf(stderr, "pgd (0/1/-1)?\n");
	fscanf(stdin, "%s", temp);

	if(atoi(temp) == 0) pgdl();
	if(atoi(temp) == 1) pgdh();
	if(atoi(temp) == -1) break;
	}

datareadable();

while(1)
	{
//	fprintf(stderr, "pgd??\n");
//	fscanf(stdin, "%s", temp);

	fprintf(stderr, "pgd=%d\n", pgdin() );
	usleep(100000);
	}

} /* end function io_test */



int hex_dump(int base, int length, int8_t *buffer, int ascii_flag)
{
int c, i, j;

if(debug_flag)
	{
	fprintf(stderr, "hex_dump(): arg base=0x%06x length=%d buffer=%p ascii_flag=%d\n", base, length, buffer, ascii_flag);
	}

if(base < 0) return 0;
if(length <= 0) return 0;
if(! buffer) return 0;

for(i = 0; i < length; i += REPORT_LENGTH)
	{
	fprintf(stderr, "0x%04x ", i + base);

	for(j = 0; j < REPORT_LENGTH; j++)
		{
		if(j == length) break;

		fprintf(stderr, "%02x ", buffer[i + j] & 0xff);		

		}

	if(ascii_flag)
		{
		fprintf(stderr, " ");

		for(j = 0; j < REPORT_LENGTH; j++)
			{	
			if(j == length) break;

			c = buffer[i + j] & 0xff;
		
			if( isalnum(c) ) fprintf(stderr, "%c", c);
			else fprintf(stderr, ".");
	
			}	

		fprintf(stderr, "\n");
		}
	}

fprintf(stderr, "\n");

return 1;
} /* end function hex_dump */



int program_chip()
{
if(debug_flag)
	{
	fprintf(stderr ,"program_chip(): arg none, using:\n");
	fprintf(stderr, " program_data=%p program_base=0x%04x, program_used=%d\n", program_data, program_base, program_used);
	fprintf(stderr, " id_data=%p id_base=0x%04x, id_used=%d\n", id_data, id_base, id_used);
	fprintf(stderr, " config_data=%p config_base=0x%04x, config_used=%d\n", config_data, config_base, config_used);
	fprintf(stderr, " eeprom_data=%p eeprom_base=0x%04x, eeprom_used=%d\n", eeprom_data, eeprom_base, eeprom_used);
	}

if(! program_data) return 0;
if(! id_data) return 0;
if(! config_data) return 0;
if(! eeprom_data) return 0;

if( (!program_used) && (!id_used) && (!config_used) && (!eeprom_used) )
	{
	fprintf(stderr, "program_chip(): nothing to program, command cancelled.\n");

	return 0;	
	}

if(program_used)
	{
	fprintf(stderr, "Writing program space.\n");

	//int write_chip(int address, int length, int8_t *buffer)
	if(! write_chip(program_base, program_used, program_data) )
		{
		fprintf(stderr, "jpp18: program_chip(): write_chip() failed for code space, programming aborted.\n");
	
		return 0;
		}
	}

if(id_used)
	{	
	fprintf(stderr, "Writing ID space.\n");

	if(! write_chip(id_base, id_used, id_data) )
		{
		fprintf(stderr, "jpp18: program_chip(): write_chip() failed for ID space, programming aborted.\n");
	
		return 0;
		}
	}

if(config_used)
	{	
	fprintf(stderr, "Writing config space.\n");

//for(i = 0; i < config_used; i++) fprintf(stderr, "i=%d data=%02x\n", i, config_data[i] & 0xff);

	if(! write_config_space(config_base, config_used, config_data) )
		{
		fprintf(stderr, "jpp18: program_chip(): write_chip() failed for config space, programming aborted.\n");
	
		return 0;
		}
	}

if(eeprom_used)
	{	
	fprintf(stderr, "Writing EEPROM space.\n");

	//int write_eeprom(int address, int length, int8_t *buffer)
	if(! write_eeprom(0, eeprom_used, eeprom_data) )
		{
		fprintf(stderr, "jpp18: program_chip(): write_eeprom() failed, programming aborted.\n");
	
		return 0;
		}
	}
	
return 1;
} /* end function program_chip */



int verify_chip()
{
int i;

if(debug_flag)
	{
	fprintf(stderr, "verify_chip(): arg none, using:\n");
	fprintf(stderr, " program_data=%p program_base=0x%04x, program_used=%d\n", program_data, program_base, program_used);
	fprintf(stderr, " id_data=%p id_base=0x%04x, id_used=%d\n", id_data, id_base, id_used);
	fprintf(stderr, " config_data=%p config_base=0x%04x, config_used=%d\n", config_data, config_base, config_used);
	fprintf(stderr, " eeprom_data=%p eeprom_base=0x%04x, eeprom_used=%d\n", eeprom_data, eeprom_base, eeprom_used);
	}

if(! program_data) return 0;
if(! id_data) return 0;
if(! config_data) return 0;
if(! eeprom_data) return 0;

if( (!program_used) && (!id_used) && (!config_used) && (!eeprom_used) )
	{
	fprintf(stderr, "Verify_chip(): nothing to verify, command cancelled.\n");

	return 0;	
	}

if(program_used)
	{
	fprintf(stderr, "Verifying program space.\n");

	// int read_chip(int address, int length, int8_t *buffer)	
	if(! read_chip(program_base, program_used, flash_read_buffer)	)
		{
		fprintf(stderr, "Could not read program space, verify cancelled.\n");
		
		return 0;
		}
	
	for(i = 0; i < program_used; i++)
		{
		if(flash_read_buffer[i] != program_data[i])
			{
			fprintf(stderr, "FAILED address=0x%04x file=%02x chip=%02x\n", i + program_base, program_data[i] & 0xff, flash_read_buffer[i] & 0xff);
			}
		}
	}

if(config_used)
	{
	fprintf(stderr, "Verifying config space.\n");
	
	// int read_chip(int address, int length, int8_t *buffer)
	if(! read_chip(config_base, config_used, flash_read_buffer) )
		{
		fprintf(stderr, "Could not read config space, verify cancelled.\n");
		
		return 0;
		}
	for(i = 0; i < config_used; i++)
		{
		/* skip non existing config locations */
		/* is this chip specific? then we need it in pic_select(); */
		
//		if( (i + config_base) == 0x300000) continue;
//		if( (i + config_base) == 0x300004) continue;
//		if( (i + config_base) == 0x300007) continue;

		if(flash_read_buffer[i] != config_data[i])
			{
			/* prevent false alerts */
			/* in those non-existing config locations the assembler sends 0xFF, and the chip reads 0x00, is this always the case?? */
			if( ( (config_data[i] & 0xff) == 0xff) && ( (flash_read_buffer[i] & 0xff) == 0x00) ) continue;

			fprintf(stderr, "FAILED address=0x%04x file=%02x chip=%02x", i + config_base, config_data[i] & 0xff, flash_read_buffer[i] & 0xff);

			/*
			NOTE!!!!
			Due to the not documented DEBUG bit in the PIC 18F22K44 this error is no error, blame Microchip:
			FAILED address=0x300006 file=41 chip=c1
			*/
			if( (i + config_base) == 0x300006)
				{
				if( (config_data[i] & 0x80 ) == 0) // bit 7 of 0x300006, the DEBUG bit, is cleared by the gpasm assembler for the 18F14K22.
					{
					if(flash_read_buffer[i] & 0x80) // bit 7 of 0x300006, the DEBUG bit, should be one in the chip to disable debug.
						{
						fprintf(stderr, "  DEBUG bit, bit 7 of 0x300006, this error can probably be ignored on PIC 18F14K22, undocumented Microchip bit, run program and see if that verifies OK.");
						}
					}
				}

			fprintf(stderr, "\n");
			}
		}
	}

if(id_used)
	{
	fprintf(stderr, "Verifying ID space.\n");
	
	// int read_chip(int address, int length, int8_t *buffer)
	if(! read_chip(id_base, id_used, flash_read_buffer) )
		{
		fprintf(stderr, "Could not read ID space, verify cancelled.\n");
		
		return 0;
		}
	for(i = 0; i < id_used; i++)
		{
		if(flash_read_buffer[i] != id_data[i])
			{
			fprintf(stderr, "FAILED address=0x%04x file=%02x chip=%02x\n", i + id_base, id_data[i] & 0xff, flash_read_buffer[i] & 0xff);
			}
		}
	}

if(eeprom_used)
	{
	fprintf(stderr, "Verifying eeprom space.\n");

	// int read_eeprom(int address, int length, int8_t *buffer)	
	if(! read_eeprom(0, eeprom_used, eeprom_read_buffer) )
		{
		fprintf(stderr, "Could not read EEPROM space, verify cancelled.\n");
		
		return 0;
		}
	for(i = 0; i < eeprom_used; i++)
		{
		if(eeprom_read_buffer[i] != eeprom_data[i])
			{
			fprintf(stderr, "FAILED address=0x%04x file=%02x chip=%02x\n", i, eeprom_data[i] & 0xff, eeprom_read_buffer[i] & 0xff);
			}
		}
	}

off();
return 1;
} /* end function verify_chip */


void print_usage()
{
fprintf(stderr,\
"\nPanteltje jppp18-%s\n\
Usage:\njppp18 [-e] [-f filename.hex] [-h] [-i filename.hex] [-p] [-t PICtype] [-v]\n\
-e            bulk erase chip.\n\
-f filename   write Intel format hex file from program space, ID space, config space, and EEPROM space, and exit.\n\
               If the filename does not end in '.hex', then '.hex' is appended to it.\n\
               This function allows cloning non-read protected PICs, the hex files can be loaded again with '-i'.\n\
-h            help (this help).\n\
-i filename   Intel format hex filename to load.\n\
               If the filename does not end in '.hex', then '.hex' is appended to it.\n\
-o            I/O test, test par port pin signals.\n\
-p            program PIC after loading hexfile, then run verify and exit, if -e is specified the PIC will be bulk erased first.\n\
-t string     PIC type.\n\
-v            verbose, prints functions and arguments.\n\
-Y            exit programming with Vdd on (default off).\n\
\n",\
PROG_VERSION);
fprintf(stderr,\
"Without either the -p or -f option specified on the command line, a menu will be presented.\n\
The -p option requires a hex file specified with -i option.\n\
Example, that will bulk erase the PIC, then program the file test.hex into the PIC, and then run verify:\n\
jppp18 -i test.hex -e -p -t 18F14K22\n\
Example, that will dump all chip areas to file test.hex, and then exit:\n\
jppp18 -f test\n\
\n"\
);

} /* end function print_usage */



int main(int argc, char **argv)
{
int a, i;
char temp[4096];
char temp2[80];
char *pic_type = 0;
int program_pic_flag;
int bulk_erase_flag;
int leave_vdd_on_flag;
FILE *bofptr;
char *hex_out_filename = 0;
struct passwd *userinfo;


/* defaults */
parport_data = PARPORT_DEFAULT_ADDRESS;

/* set a pic type */
pic_type = strsave("18F14K22");
if(! pic_type)
	{
	fprintf(stderr, "jppp18: could not allocate space for pic_type, aborting.\n");
	
	exit(1);
	}

program_pic_flag = 0;
bulk_erase_flag = 0;
leave_vdd_on_flag = 0;

/* end defaults */


/* get user info */
errno = 0;
userinfo = getpwuid(getuid() );
if(! userinfo)
	{
	fprintf(stderr, "jppp18: main(): getpwuid() failed because: ");
	perror("");
	fprintf(stderr, "jppp18: aborting.\n");	

	exit(1);	
	}


/* test if run as root */
if(strcmp(userinfo -> pw_dir, "/root") != 0)
	{
	fprintf(stderr, "jppp18: you need to be root to run this program, perhaps try sudojppp18, aborting.\n");
	
	exit(1);
	}
                

/* calculate par port addresses */
parport_status = parport_data + 1;
parport_control = parport_data + 2;

/* allow I/O to parport */
if( iopl(3) == -1)
	{
	fprintf(stderr, "jppp18: main(): iopl(3) faled because: ");
	perror("");	
	fprintf(stderr, "jppp18: aborting.\n");
	
	exit(1);
	}


/* all pins low */
allpinslow();

/* proces any command line arguments */
while(1)
	{
	a = getopt(argc, argv, "ef:hi:opt:vY");
	if(a == -1) break;

	switch(a)
		{
		case 'e': // bulk erase
			bulk_erase_flag = 1;
			break;
		case 'f':	// f write Intel format hex file from program space, ID space, config space, and EEPROM space.
			hex_out_filename = strsave(optarg);
			if(! hex_out_filename)
				{
				fprintf(stderr, "jppp18: could not allocate space for hex_out_filename, aborting.\n");

				exit(1);
				}
			break;
		case 'h': // help
			print_usage();
			exit(1);
			break;
		case 'i': // Intel format hex file
			hex_in_filename = strsave(optarg);
			if(! hex_in_filename)
				{
				fprintf(stderr, "jppp18: could not allocate space for hex_in_filename, aborting.\n");
				
				exit(1);
				}
			break;
		case 'o':	// I/O test
			io_test();
			break;
		case 'p':
			program_pic_flag = 1;
			break;
		case 't': // PIC type
			if(pic_type) free(pic_type);
			pic_type = strsave(optarg);
			if(! pic_type)
				{
				fprintf(stderr, "jppp18: could not allocate space for pic_type, aborting.\n");
				
				exit(1);
				}
			break;			
		case 'v': // verbose
			debug_flag = 1;
			break;
		case 'Y': // exit with Vdd on
			leave_vdd_on_flag = 1;
			break;
        case -1:
        	break;
		case '?':
			if (isprint(optopt) )
 				{
 				fprintf(stderr, "hxl: unknown option `-%c'.\n", optopt);
 				}
			else
				{
				fprintf(stderr, "hxl: unknown option character `\\x%x'.\n", optopt);
				}
			print_usage();

			exit(1);
			break;			
		default:
			print_usage();

			exit(1);
			break;
		}/* end switch a */
	}/* end while getopt() */

/* select a pic */
/* this will allocate space for buffers too */
if(! pic_select(pic_type) )
	{
	fprintf(stderr, "jppp18: pic_select failed, aborting.\n");

	exit(1);
	}

/* if a hex output is requested write it and exit */
if(hex_out_filename)
	{
	// int write_hex_file(char *filename);
	if(! write_hex_file(hex_out_filename) )
		{
		fprintf(stderr, "write_hex_file failed.\n");
		}

	if(leave_vdd_on_flag) vdd5();

	exit(0);
	}

/* program the PIC if enough info on the command line */
/*
NOTE:
will have to check for presence of PIC type later.
*/

if(hex_in_filename)
	{
	strcpy(temp, hex_in_filename);
	
	if(strstr(temp, ".hex") != (temp + strlen(temp) - 4) ) strcat(temp, ".hex");

	if(! load_hex_file(temp) )
		{
		fprintf(stderr, "jppp18: load_hex_file failed, aborting.\n");
		
		exit(1);
		}
	}	


if(bulk_erase_flag)
	{
	bulk_erase();
	
	if(! program_pic_flag) exit(0);
	}


if(program_pic_flag)
	{
	if(! hex_in_filename)
		{
		fprintf(stderr, "jppp18: no hex file specified, use the -i flag, aborting.\n");

		exit(1);
		}

	if(! program_chip() )
		{
		fprintf(stderr, "jppp18: program_chip failed, aborting.\n");
	
		exit(1);
		}
	
	if(! verify_chip() )
		{
		fprintf(stderr, "jppp18: verify chip failed, aborting.\n");

		exit(1);
		}

	fprintf(stderr, "Ready.\n");

	if(leave_vdd_on_flag) vdd5();

	exit(0);
	} /* end if program_pic_flag */
	

/* menu */
while(1)
	{
	allpinslow();

	/* display menu */
	fprintf(stderr, "\njppp18-%s Copyright (c) Jan Panteltje 2010-always.\n", PROG_VERSION);
	fprintf(stderr, "b dump code space as binary to file.\n");
	fprintf(stderr, "c hex dump config area.\n");
	fprintf(stderr, "d hex dump ID area.\n");
	fprintf(stderr, "e erase chip.\n");
	fprintf(stderr, "f write Intel format hex file from program space, ID space, config space, and EEPROM space.\n");
	fprintf(stderr, "h hex dump program area.\n");
	fprintf(stderr, "i load Intel format hex file.\n");
	fprintf(stderr, "m hex dump eeprom.\n");
    fprintf(stderr, "o I/O test, test par port signals.\n");
	fprintf(stderr, "p program chip.\n");
	fprintf(stderr, "t PIC type select.\n");
	fprintf(stderr, "v verify chip.\n");
	fprintf(stderr, "x exit.\n");
	fprintf(stderr, "Y leave Vdd on.\n");
	fprintf(stderr, "y leave Vdd off.\n");
	fprintf(stderr, "z print device ID.\n");
	fprintf(stderr, " Select\n");

	/* get user selection */	
	fscanf(stdin, "%s", temp);
	switch(temp[0])
		{
		case 'b': // dump code space as binary to stdout.
			fprintf(stderr, "filename?\n");
			fscanf(stdin, "%s", temp);
			bofptr = fopen(temp, "r");
			if(bofptr)
				{
				fclose(bofptr);

				fprintf(stderr, "file %s already exists, overwrite (y/n)?\n", temp);
				fscanf(stdin, "%s", temp2);
				if( toupper(temp2[0]) != 'Y')
					{
					fprintf(stderr, "dump code space to file cancelled.\n");

					break;
					}				
				}

			bofptr = fopen(temp, "w");
			if(! bofptr)
				{
				fprintf(stderr, "could not open file %s for write, command cancelled.\n", temp);
				
				break;
				}

			fprintf(stderr, "reading code space.\n");			
			if(! read_chip(program_base, program_size, flash_read_buffer) )
				{
				fprintf(stderr, "read_chip failed.\n");

				break;
				}

			fprintf(stderr, "writing to file %s.\n", temp);
			for(i = 0; i < program_size; i++)
				{
				fprintf(bofptr, "%c", flash_read_buffer[i]);
				}	

			fclose(bofptr);
			break;	
		case 'c':	// hex dump config area
			// int read_chip(int address, int length);
			if(! read_chip(config_base, config_size, flash_read_buffer) )
				{
				fprintf(stderr, "read_chip failed.\n");
				break;
				}

			hex_dump(config_base, config_size, flash_read_buffer, 0); 
			break;	
		case 'd': // hex dump ID area
			// int read_chip(int address, int length);
			if(! read_chip(id_base, id_size, flash_read_buffer) )
				{
				fprintf(stderr, "read_chip failed.\n");

				break;
				}

			hex_dump(id_base, id_size, flash_read_buffer, 0); 
			break;	
		case 'e':	// erase chip
			fprintf(stderr, "bulk erase chip (y/n)?\n");	
			fscanf(stdin, "%s", temp);
			if(temp[0] != 'y')
				{
				fprintf(stderr, "Erase command cancelled.\n");

				break;
				} 

			if(! bulk_erase() )
				{
				fprintf(stderr ,"bulk_erase failed.\n");			
				break;
				}

			fprintf(stderr, "erasure complete.\n");
			break;	
		case 'f':	// write Intel format hex file from program space, ID space, config space, and EEPROM space
			fprintf(stderr, "filename?\n");
			fscanf(stdin, "%s", temp);

			// int write_hex_file(char *filename);
			if(! write_hex_file(temp) )
				{
				fprintf(stderr, "write_hex_file failed.\n");
				}
			
			break;
		case 'h':	// hex dump program area
			// int read_chip(int address, int length);
			if(! read_chip(program_base, program_size, flash_read_buffer) )
				{
				fprintf(stderr, "read_chip failed.\n");

				break;
				}

			hex_dump(program_base, program_size, flash_read_buffer, 1); 
			break;	
		case 'i':	// load hex file
			fprintf(stderr, "hex filename?\n");
			if(hex_in_filename) free(hex_in_filename);
			a = fscanf(stdin, "%as", &hex_in_filename);
			if( (a == EOF) || (a == 0) )
				{
				fprintf(stderr, "jppp18: could not read hex_in_filename, aborting.\n");

				exit(1);
				}

			strcpy(temp, hex_in_filename);
	
			if(strstr(temp, ".hex") != (temp + strlen(temp) - 4) ) strcat(temp, ".hex");

			if(! load_hex_file(temp) )
				{
				fprintf(stderr, "load_hex_file failed.\n");
				break;
				}
			break;	
		case 'm': // hex dump eeprom
			if(! read_eeprom(0, eeprom_size, eeprom_read_buffer) )
				{
				fprintf(stderr, "read_eeprom failed.\n");

				break;
				}

			hex_dump(0, eeprom_size, eeprom_read_buffer, 1); 
			break;	
		case 'o':// I/O test
			io_test();
			break;
		case 'p': //  program chip
			if(! hex_in_filename)
				{
				fprintf(stderr, "load a hex file first.\n");

				break;
				}

			if(! program_chip() )
				{
				fprintf(stderr, "program_chip failed.\n");

				break;
				}

			if(! verify_chip() )
				{
				fprintf(stderr, "verify chip failed.\n");
				break;
				}

			fprintf(stderr, "verify chip ready.\n");
			break;	
		case 't':	// pic select
			fprintf(stderr, "PIC type?\n");
			if(pic_type) free(pic_type);
			a = fscanf(stdin, "%as", &pic_type);
			if( (a == EOF) || (a == 0) ) 
				{
				fprintf(stderr, "jppp18: could not read pic_type, aborting.\n");
				exit(1);
				}

			if(! pic_select(pic_type) )
				{
				fprintf(stderr, "unknown PIC type.\n");
				}

			/* clear data from any already loaded hex file */
			eeprom_used = 0;
			config_used = 0;
			id_used = 0;
			program_used = 0;
			if(hex_in_filename)
				{
				free(hex_in_filename);
				hex_in_filename = 0;
				}

			break;
		case 'v':	// verify chip
			if(! hex_in_filename)
				{
				fprintf(stderr, "load a hex file first.\n");

				break;
				}
		
			verify_chip();
			break;
		case 'x': // exit
			fprintf(stderr, "jppp18: User abort.\n");
			exit(0);
			break;	
		case 'Y':	// Vdd on
			leave_vdd_on_flag = 1;
			vdd5();
			break;
		case 'y':	// Vdd off
			leave_vdd_on_flag = 0;
			vdd0();
			break;
		case 'z': // print device ID
			// int read_chip(int address, int length);
			if(! read_chip(devid_base, devid_size, flash_read_buffer) )
				{
				fprintf(stderr, "read_chip failed.\n");

				break;
				}

			hex_dump(devid_base, devid_size, flash_read_buffer, 0); 
			break;
		default:
			fprintf(stderr, "Unknown choice '%c'.\n", temp[0]);
			break;
		} /* end switch choice */
	
	off();
	if(leave_vdd_on_flag) vdd5();
	} /* end while menu */

exit(1);
} /* end function main */



