
/*
jppp18 (c) 2010-always Jan Panteltje

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



                                            /* set tabs to 4 to read this source */



#include "jppp18.h"


void delay(int k)
{ 
int i;
int a;

if(debug_flag)
	{
	fprintf(stderr, "delay(): arg k=%d\n", k);
	}
	

for(i = 0; i < k; i++) {a = i + 1; }

return;
} /* end function delay */


void on()
{
if(debug_flag)
	{
	fprintf(stderr, "on(): arg none\n");
	}

pgdl();
pgcl();

vdd5();
usleep(100000);
vpp9();
}


void off()
{
if(debug_flag)
	{
	fprintf(stderr, "off(): arg none\n");
	}

pgdl();
pgcl();

vpp0();
usleep(100000);
vdd0();
}


int rcvdata()				// read data, sends command 9, tdelay P5, 8 dummy clocks with data low, delay P6, P14 data to clock setup, 8 clocks to clock out byte.
{
int i;
int b;
int w = 0;

if(debug_flag)
	{
	fprintf(stderr, "rcvdata(): arg none\n");
	}
	
datareadable();				// SLCTIN up for pull-up
delay(2);

/* read 8 dummy bits */
pgdl();
for(i = 0; i < 8; i++)
	{
	pgch();
	delay(1);

	pgcl();
	delay(1);
	}

delay(1);

/* read 8 data bits */
for(i = 0; i < 8; i++)
	{
	pgch();
	delay(1);

	b = pgdin();

	pgcl();
	delay(1);
	
	w = w | (b << i);		// 8 data bits
	}

delay(1);
pgdl();

if(debug_flag)
	{
	fprintf(stderr, "rcvdata(): returning w=%04x\n", w);
	}

return w;
}
 

void snddata(int w)			// Sends 16-bit word from bottom of w
{
int i;

if(debug_flag)
	{
	fprintf(stderr, "snddata(): arg w=%02x\n", w);
	}
	
datawritable();
delay(2);

for (i = 16; i > 0; i--)
	{
	pgch();
	delay(1);

	if(w & 1) pgdh();
	else pgdl();

	delay(1);

	pgcl();	
	delay(1);

	w = w >> 1;
	}

pgdl();

delay(2);
}
 

void sndcmd(int b)			// Sends 4-bit command from bottom of b
{
int i;

if(debug_flag)
	{
	fprintf(stderr, "sndcmd(): arg b=%02x\n", b);
	}
	
datawritable();
delay(2);

for (i = 4; i > 0; i--)
	{
	pgch();
	delay(1);

	if(b & 1) pgdh();
	else pgdl();
	delay(1);
	
	pgcl();
	delay(1);

	b = b >> 1;
	}

pgdl();

delay(2);
}
 

void sndcmd_prog(int congifuration_bits_mode_flag)			// Sends 4-bit nop command and keeps clock low for 1 ms, or 5 ms for programming configuration bits */
{
int i;

if(debug_flag)
	{
	fprintf(stderr, "sndcmd_prog(): arg congifuration_bits_mode_flag=%d\n", congifuration_bits_mode_flag);
	}

datawritable();
delay(2);

/* data line to zero */
pgdl();

/* 3 clocks with normal delay */
for (i = 0; i < 3; i++)
	{
	pgch();
	delay(1);

	pgcl();
	delay(1);
	}

/* clock to one for bit 4 */
pgch();

/* clock remains high for P9 = 1 ms,  P9A = 5 ms clock high for programming configuration bits */
if(congifuration_bits_mode_flag) usleep(5000);
else usleep(1000);

pgcl();
/* clock remains low for P10 = 100 us */
usleep(100);

/* write 16 zeros */
for(i = 0; i < 16; i++)
	{
	pgch();
	delay(1);

	pgcl();
	delay(1);
	}
} /* end function sndcmd_prog */


void snd_eeprom_data(int w)			// Sends 8-bit word from bottom of w
{
int i;

if(debug_flag)
	{
	fprintf(stderr, "snd_eeprom_data(): arg w=%02x\n", w);
	}
	
datawritable();
delay(2);

/* first 8 clocks with data low */
pgdl();

for(i = 0; i < 8; i++)
	{
	pgch();
	delay(1);

	pgcl();	
	delay(1);
	}
	
/* send byte */
for (i = 8; i > 0; i--)
	{
	pgch();
	delay(1);

	if(w & 1) pgdh();
	else pgdl();

	delay(1);

	pgcl();	
	delay(1);

	w = w >> 1;
	}

pgdl();

delay(2);
}

 

int bulk_erase()
{
if(debug_flag)
	{
	fprintf(stderr, "bulk_erase(): arg none\n");
	}

fprintf(stderr, "Erasing chip.\n");

/*
BULKERASE OPTIONS
Description						Data (3C0005h:3C0004h
Chip Erase						0F8Fh
Erase User IDs					0088h
Erase Data EEPROM				0084h
Erase Boot Block				0081h
Erase Config Bits				0082h
Erase Program Flash Block		00180h
Erase Program Flash Block		10280h
Erase Program Flash Block		20480h
Erase Program Flash Block		30880h
*/

on();

/*
bulk erase
4 bit command, command payload, core instruction
*/
// 0000   0E 3C  MOVLW 3Ch
sndcmd(0x0);
snddata(0x0e3c);

// 0000   6E F8  MOVWF TBLPTRU
sndcmd(0);
snddata(0x6ef8);

// 0000   0E 00  MOVLW 00h
sndcmd(0);
snddata(0x0e00);

// 0000   6E F7  MOVWF TBLPTRH
sndcmd(0);
snddata(0x6ef7);

// 0000   0E 05  MOVLW 05h
sndcmd(0);
snddata(0x0e05);

// 0000   6E F6  MOVWF TBLPTRL
sndcmd(0);
snddata(0x6ef6);

// 1100   0F 0F  Write 0Fh to 3C0005h
sndcmd(12);
snddata(0x0f0f);

// 0000   0E 3C  MOVLW 3Ch
sndcmd(0);
snddata(0x0e3c);

// 0000   6E F8  MOVWF TBLPTRU
sndcmd(0);
snddata(0x6ef8);

// 0000   0E 00  MOVLW 00h
sndcmd(0);
snddata(0x0e00);

// 0000   6E F7  MOVWF TBLPTRH
sndcmd(0);
snddata(0x6ef7);

// 0000   0E 04  MOVLW 04h
sndcmd(0);
snddata(0x0e04);

// 0000   6E F6  MOVWF TBLPTRL
sndcmd(0);
snddata(0x6ef6);

// Write 8F8Fh TO 3C0004h to erase entire device.
// 1100   8F 8F
sndcmd(12);
snddata(0x8f8f);

// 0000   00 00  NOP
sndcmd(0);
snddata(0x0000);

// Hold PGD low until erase completes.
// 0000   00 00
sndcmd(0);
snddata(0x0000);

sleep(1);

off();

return 1;
} /* end function bulk_erase */



int read_chip(int address, int length, int8_t *buffer)
{
int i;
int value;

if(debug_flag)
	{
	fprintf(stderr, "read_chip(): arg address=0x%04x length=%d, buffer=%p\n", address, length, buffer);
	}

if(address < 0) return 0;
if(length < 0) return 0;
if(! buffer) return 0;

on();

// 0x000000 - 0x00ffff Program FLASH
// 0x200000 - 0x200007 User ID registers.
// 0x300001 - 0x30000d configuration bits 
// 0x3ffffe - 0x3fffff device ID 


// address U
sndcmd(0);
value = 0xe00 | ( (address & 0xff0000) >> 16);		// MOVLW <Addr[23:16]
snddata(value);

sndcmd(0);
snddata(0x6ef8);									// MOVWF TBLPTRU


// address H
sndcmd(0);
value = 0xe00 | ( (address & 0x00ff00) >> 8);		// MOVLW <Addr[15:8]
snddata(value);

sndcmd(0);
snddata(0x6ef7);									// MOVWF TBLPTRH


// address L
sndcmd(0);
value = 0xe00 | (address & 0x0000ff);				// MOVLW <Addr[7:0]
snddata(value);

sndcmd(0);
snddata(0x6ef6);									// MOVWF TBLPTRL

for(i = 0; i < length; i++)
	{
	// read and post increment
	sndcmd(9);										// TBLRD *+
	buffer[i] = rcvdata();							// get data
	}

off();

return 1;
} /* end function read_chip */



int write_chip(int address, int length, int8_t *buffer)
{
int i, j;
int value;
int write_position;
int byte_count;
int buffer_position;

if(debug_flag)
	{
	fprintf(stderr, "write_chip(): arg address=0x%04x length=%d, buffer=%p\n", address, length, buffer);
	}

if(address < 0) return 0;
if(length < 0) return 0;
if(! buffer) return 0;

if(length % 2)
	{
	fprintf(stderr, "write_flash(): length must be even, command cancelled.\n");

	return 0;
	}

on();

// Step 1: Direct access to program Flash.
// 0000               8E A6           BSF     EECON1, EEPGD
sndcmd(0);
snddata(0x8ea6);


// 0000               9C A6           BCF     EECON1, CFGS
sndcmd(0);
snddata(0x9ca6);

// 0000               84 A6           BSF     EECON1, WREN
sndcmd(0);
snddata(0x84a6);

//fprintf(stderr, "address=0x%04x length=%d\n", address, length);

write_position = address;
byte_count = 0;
/* for all bytes in buffer */
for(i = 0; i < length; i += 16)
	{
	/* initialize write for 16 bytes to write buffer */
	// Step 2: Point to row to write.
	// 0000          0E <Addr[21:16]>     MOVLW   <Addr[21:16]>
	sndcmd(0);
	value = 0xe00 | ( (write_position & 0xff0000) >> 16);
	snddata(value);

	// 0000               6E F8           MOVWF   TBLPTRU
	sndcmd(0);
	snddata(0x6ef8);

	// 0000          0E <Addr[15:8]>      MOVLW   <Addr[15:8]>
	sndcmd(0);
	value = 0xe00 | ( (write_position & 0x00ff00) >> 8);
	snddata(value);

	// 0000               6E F7           MOVWF   TBLPTRH
	sndcmd(0);
	snddata(0x6ef7);

	// 0000           0E <Addr[7:0]>      MOVLW   <Addr[7:0]>
	sndcmd(0);
	value = 0xe00 | (write_position & 0x0000ff);
	snddata(value);

	// 0000               6E F6           MOVWF   TBLPTRL
	sndcmd(0);
	snddata(0x6ef6);

	fprintf(stderr, "0x%04x\r", i);

	/* fill 16 bytes write buffer */
	/* first 14 bytes */
	buffer_position = 0;
	for(j = 0; j < 14; j += 2)
		{
		// Step 3: Load write buffer. Repeat for all but the last two bytes.
		
		if(byte_count < (length - 2) )
			{
			// 1101           <MSB><LSB>          Write 2 bytes and post-increment address by 2.
			sndcmd(13);
// WASR
			snddata(  (  (buffer[i + buffer_position + 1] & 0xff) << 8) | (buffer[i + buffer_position] & 0xff) );
//			snddata(  (  (buffer[i + buffer_position ] & 0xff) << 8) | (buffer[i + buffer_position + 1] & 0xff) );

//fprintf(stderr, "writing A: write_position=%d buffer_position=%d j=%d buffer[%d]=0x%02x\n", write_position, buffer_position,  j, i + buffer_position, (buffer[i + buffer_position] & 0xff) ); 
//fprintf(stderr, "writing A: write_position=%d buffer_position=%d j=%d buffer[%d]=0x%02x\n", write_position, buffer_position + 1, j, i + buffer_position + 1, (buffer[i + buffer_position + 1] & 0xff) ); 

			buffer_position += 2;
			write_position += 2;
			byte_count += 2;
			}			
		else
			{
			break;
			}
		}

	// To continue writing data, repeat steps 2 through 4, where the Address Pointer is incremented by 2 at each iteration of the loop.

	/* write bytes 15 and 16, keep pgc high after last byte for 1 mS */
	// Step 4: Load write buffer for last two bytes and start programming.
	// 1111           <MSB><LSB>          Write 2 bytes and start programming.
	sndcmd(15);
//WASR
	snddata( (  (buffer[i + buffer_position + 1] & 0xff ) << 8) | ( buffer[i + buffer_position]  & 0xff) );
//	snddata( (  (buffer[i + buffer_position ] & 0xff ) << 8) | ( buffer[i + buffer_position + 1]  & 0xff) );

//fprintf(stderr, "writing B: write_position=%d buffer_position=%d j=%d buffer[%d]=0x%02x\n", write_position, buffer_position,  j, i + buffer_position, (buffer[i + buffer_position] & 0xff) ); 
//fprintf(stderr, "writing B: write_position=%d buffer_position=%d j=%d buffer[%d]=0x%02x\n", write_position, buffer_position + 1,  j, i + buffer_position + 1, (buffer[i + buffer_position + 1] & 0xff) ); 

	sndcmd_prog(0); /* forces 1 ms delay for normal program FLASH */

	write_position += 2;

	byte_count += 2;
	if(byte_count >= length) break;

	} /* end for i */

usleep(100000);

off();

return 1;
} /* end function write_chip */


	
int write_config_space(int address, int length, int8_t *buffer)
{
int i;
int value;
int write_position;

if(debug_flag)
	{
	fprintf(stderr, "write_config_space(): arg address=0x%04x length=%d, buffer=%p\n", address, length, buffer);
	}

if(address < 0) return 0;
if(length < 0) return 0;
if(! buffer) return 0;

if(length % 2)
	{
	fprintf(stderr, "write_config_space(): length must be even, command cancelled.\n");

	return 0;
	}

on();

// Step 1: Direct access to program Flash.
// 0000               8E A6           BSF     EECON1, EEPGD
sndcmd(0);
snddata(0x8ea6);

// 0000               8C A6           BSF EECON1, CFGS
sndcmd(0);
snddata(0x8ca6);

// 0000               84 A6           BSF     EECON1, WREN
sndcmd(0);
snddata(0x84a6);

//fprintf(stderr, "address=0x%04x length=%d\n", address, length);


/* for all bytes in buffer */
write_position = address;	// points to first byte in buffer, even position
for(i = 0; i < length; i++)
	{
	// Step 2(1): Set Table Pointer for configuration byte to be written. Write even/odd addresses.
	//	0000                         0E 30               MOVLW 30h
	sndcmd(0);
	value = 0xe00 | ( (write_position & 0xff0000) >> 16);
	snddata(value);

	// 0000                         6E F8               MOVWF TBLPTRU
	sndcmd(0);
	snddata(0x6ef8);

	// 0000                         0E 00               MOVLW 00h
	sndcmd(0);
	value = 0xe00 | ( (write_position & 0x00ff00) >> 8);
	snddata(value);

	// 0000                         6E F7               MOVWF TBLPRTH
	sndcmd(0);
	snddata(0x6ef7);


	// 0000                         0E 00               MOVLW 00h			set EVEN address  // send LSB
	sndcmd(0);
	value = 0xe00 | (write_position & 0x0000ff);
	snddata(value);

	//0000                         6E F6               MOVWF TBLPTRL
	sndcmd(0);
	snddata(0x6ef6);

	/* data */
	if(i % 2) // odd LSB ignored
		{
//		fprintf(stderr, "Writing odd config i=%d address=0x%06x data=0x%04x\n", i, write_position, 0x0000 |  ( (buffer[i] & 0xff ) << 8) ); 

		// 1111                <MSB><LSB ignored>           Load 2 bytes and start programming.	ODD
		sndcmd(15);
		snddata( 0x0000 |  ( (buffer[i] & 0xff ) << 8) );			// use MSB
		}
	else	//	 even MSB ignored
		{ 

		/* bit 7 CONFIG4L 300006h must be 1 to disable debug!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

		if(write_position == 0x300006) buffer[i] |= 0x80; /* set !DEBUG bit 7 in buffer, so verify won't complain later */

//		a = 0x80; /* set !DEBUG bit 7 */
//		else a = 0;
		

//		fprintf(stderr, "Writing even config i=%d address=0x%06x data=0x%04x\n", i, write_position, 0x0000 | ( buffer[i] & 0xff) );

		// 1111                <MSB ignored><LSB>           Load 2 bytes and start programming. EVEN
		sndcmd(15);
		snddata( 0x0000 | ( buffer[i] & 0xff)  );						// use LSB
		}

	// 0000                         00 00               NOP - hold PGC high for time P9A and low for time P10.
	sndcmd_prog(1);	// 5 ms delay

	write_position++;
	} /* end for i */

// Note 1:     Enabling the write protection of Configuration bits (WRTC = 0 in CONFIG6H) will prevent further writing of Configuration bits.
// Always write all the Configuration bits before enabling the write protection for Configuration bits.

usleep(100000);

off();

return 1;
} /* end function write_config_space */




int read_eeprom(int address, int length, int8_t *buffer)
{
int i;

if(debug_flag)
	{
	fprintf(stderr, "read_eeprom(): arg address=0x%04x length=%d buffer=%p\n", address, length, buffer);  
	}

if(address < 0) return 0;
if(address > 255) return 0;
if( (address + length) > 256) return 0;
if(length < 0) return 0;
if(! buffer) return 0;

on();

// Step 1: Direct access to data EEPROM.
// 0000                    9E A6             BCF EECON1, EEPGD
sndcmd(0);
snddata(0x9ea6);

// 0000                    9C A6             BCF EECON1, CFGS
sndcmd(0);
snddata(0x9ca6);

for(i = address; i < (address + length); i++)
	{
	// Step 2: Set the data EEPROM Address Pointer.
	// 0000                  0E <Addr>           MOVLW <Addr>
	sndcmd(0);
	snddata(0x0e00 | (i & 0xff) );

	// 0000                    6E A9             MOVWF EEADR
	sndcmd(0);
	snddata(0x6ea9);

	// Step 3: Initiate a memory read.
	// 0000                    80 A6             BSF EECON1, RD
	sndcmd(0);
	snddata(0x80a6);

	// Step 4: Load data into the Serial Data Holding register.
	// 0000                    50 A8             MOVF EEDATA, W, 0
	sndcmd(0);
	snddata(0x50a8);

	// 0000                    6E F5             MOVWF TABLAT
	sndcmd(0);
	snddata(0x6ef5);

	// 0000                    00 00             NOP
	sndcmd(0);
	snddata(0x0000);

	// 0010                <MSB><LSB>            Shift Out Data(1)
	// Note 1:       The <LSB> is undefined. The <MSB> is the data.
	sndcmd(2);									// Shift out TABLAT register
	buffer[i] = rcvdata();							// get data
	}

off();
return 1;
} /* end function read_eeprom */


int write_eeprom(int address, int length, int8_t *buffer)
{
int i;
int8_t a;
int loop_count;

if(debug_flag)
	{
	fprintf(stderr, "write_eeprom(): arg address=0x%04x length=%d data=%p\n", address, length, buffer);  
	}

if(address < 0) return 0;
if( (address + length) > 256) return 0;
if(length < 0) return 0;
if(length > 256) return 0;
if(! buffer) return 0;

on();

// Step 1: Direct access to data EEPROM.
// 0000                    9E A6             BCF EECON1, EEPGD
sndcmd(0);
snddata(0x9ea6);

// 0000                    9C A6             BCF EECON1, CFGS
sndcmd(0);
snddata(0x9ca6);

for(i = address; i < (address + length); i++)
	{
	// Step 2: Set the data EEPROM Address Pointer.
	// 0000                0E <Addr>           MOVLW <Addr>
	sndcmd(0);
	snddata(0x0e00 | (i & 0xff) );

	// 0000                  6E A9             MOVWF EEADR
	sndcmd(0);
	snddata(0x6ea9);

	// Step 3: Load the data to be written.
	// 0000                0E <Data>           MOVLW <Data>
	sndcmd(0);
	snddata(0x0e00 | (buffer[i] & 0xff) );

	// 0000                  6E A8             MOVWF EEDATA
	sndcmd(0);
	snddata(0x6ea8);

	// Step 4: Enable memory writes.
	// 0000                  84 A6             BSF EECON1, WREN
	sndcmd(0);
	snddata(0x84a6);

	// Step 5: Initiate write.
	// 0000                  82 A6             BSF EECON1, WR
	sndcmd(0);
	snddata(0x82a6);

	// 0000                  00 00             NOP
	sndcmd(0);
	snddata(0x0000);

	// 0000                  00 00             NOP					write starts on 4th clock of this instruction
	sndcmd(0);
	snddata(0x0000);

	// Step 6: Poll WR bit, repeat until the bit is clear.
	/* make sure we have a safe timeout, to get Vpp back to zero */
	loop_count = 0;
	while(1)
		{
		// 0000                  50 A6             MOVF EECON1, W, 0
		sndcmd(0);
		snddata(0x50a6);	

		// 0000                  6E F5             MOVWF TABLAT
		sndcmd(0);
		snddata(0x6ef5);


		// ???????????????
		// 0010                <MSB><LSB>            Shift Out Data(1)
		// Note 1:       The <LSB> is undefined. The <MSB> is the data.

		// EECON1
		// bit 1 WR: Write Control bit
		// 1 = Initiates a data EEPROM erase/write cycle or a program memory erase cycle or write cycle.
		// (The operation is self-timed and the bit is cleared by hardware once write is complete.
		// The WR bit can only be set (not cleared) by software.)
		// 0 = Write cycle to the EEPROM is complete

		sndcmd(2);							// Shift out TABLAT register
		a = rcvdata();						// get data

//		fprintf(stderr, "i=%d a=%02x\n", i, a & 0xff);

		if(! (a & 0x02) ) break;		// if bit 1 is clear, then this EEPROM location is programmed. 

		/* test for timeout */
		loop_count++;
		if(loop_count == 1000)
			{
			fprintf(stderr, "write_eeprom(): timeout in EEPROM programming location 0x%06x, EEPROM programming cancelled.\n", i + eeprom_base);
			
			off();
			return 0;
			}

		} /* end while poll bit 1 EECON1 for ready */

	// 0010               <MSB><LSB>           Shift out data(1)
	sndcmd(2);
	snd_eeprom_data(buffer[i] & 0xff);

	// Step 7: Hold PGC low for time P10 (100us).
	usleep(100);

	// Step 8: Disable writes.
	// 0000                  94 A6             BCF EECON1, WREN
	sndcmd(0);
	snddata(0x94a6);

	// Repeat steps 2 through 8 to write more data.
    }                                                                      

off();
return 1;
} /* end function write_eeprom */



